from __future__ import absolute_import
import numpy as np
import base64


# TODO: add date/time/datetime serialiser

_serialisers = set()


def register_serialiser(cls):
    global _serialisers

    _serialisers.add(cls())


def serialisers():
    return _serialisers


def find_serialiser(python_type):
    for s in _serialisers:
        for t in s.python_types:
            if isinstance(python_type, t):
                return s


def find_deserialiser(serialised_type):
    for s in _serialisers:
        if serialised_type in s.serialised_types:
            return s


class SerialiserMetaClass(type):
    def __new__(cls, clsname, bases, attrs):
        '''Automatically registers Serialiser subclasses at class definition time.
        '''
        newclass = super(SerialiserMetaClass, cls).__new__(cls, clsname, bases, attrs)
        register_serialiser(newclass)  # here is your register function
        return newclass


class Serialiser(object):
    __metaclass__ = SerialiserMetaClass

    python_types = tuple()
    serialised_types = tuple()

    def to_json(self, obj):
        raise ValueError('Unknown data type: {}'.format(type(obj)))

    def from_json(self, jobj):
        raise ValueError('Unknown data type: {}'.format(jobj))


class NumpySerialiser(Serialiser):
    python_types = (np.ndarray, np.generic)
    serialised_types = ('ndarray', 'npgeneric')

    def to_json(self, obj):
        if isinstance(obj, np.ndarray):
            return {
                '__type__': 'ndarray',
                'data': base64.b64encode(obj.tostring()),
                'dtype': obj.dtype.str,
                'shape': obj.shape,
            }
        elif isinstance(obj, (np.bool_, np.number)):
            return {
                '__type__': 'npgeneric',
                'data': base64.b64encode(obj.tostring()),
                'dtype': obj.dtype.str,
            }

        return super(NumpySerialiser, self).to_json(obj)

    def from_json(self, jobj):
        obj = np.fromstring(
            base64.b64decode(jobj['data']),
            dtype=np.dtype(jobj['dtype'])
        )
        if jobj.get('__type__') == 'ndarray':
            return obj.reshape(jobj['shape'])
        if jobj.get('__type__') == 'npgeneric':
            return obj[0]

        return super(NumpySerialiser, self).from_json(jobj)


class PythonTypeSerialiser(Serialiser):
    python_types = (set, tuple, complex)
    serialised_types = ('set', 'tuple', 'complex')

    def to_json(self, obj):
        if isinstance(obj, set):
            return {
                '__type__': 'set',
                'data': list(obj),
            }
        if isinstance(obj, tuple):
            return {
                '__type__': 'tuple',
                'data': list(obj)
            }
        if isinstance(obj, complex):
            return {
                '__type__': 'complex',
                'data': obj.__repr__()
            }

        return super(PythonTypeSerialiser, self).to_json(obj)

    def from_json(self, jobj):
        obj = np.fromstring(
            base64.b64decode(jobj['data']),
            dtype=np.dtype(jobj['dtype'])
        )
        if jobj.get('__type__') == 'set':
            return set(obj['data'])
        if jobj.get('__type__') == 'tuple':
            return tuple(obj['data'])
        if jobj.get('__type__') == 'complex':
            return complex(obj['data'])

        return super(PythonTypeSerialiser, self).from_json(jobj)



"""Provides Object -> JSON -> Object serialisation functionality.

This code is designed to avoid any `eval` calls which could be
exploited, if the database were compromised, with malicious code.

To avoid calling `eval` on the class type, classes must be registered
with the `register_type` function.

Note: A badly written class could still parse internal data using eval.
"""

from . import serialisable
from . import serialiser
from .serialisable import Serialisable
from .serialiser import Serialiser
import json
from json import *


def to_dict(obj):
    s = serialiser.find_serialiser(obj)
    if s:
        return s.to_json(obj)

    raise TypeError('Unable to serialise object of type {}'.format(type(obj)))


def from_dict(jobj):
    if '__type__' in jobj:
        s = serialiser.find_deserialiser(jobj['__type__'])
        if s:
            return s.from_json(jobj)
    return jobj


def load(*args, **kwargs):
    kwargs['object_hook'] = from_dict
    return json.load(*args, **kwargs)


def loads(*args, **kwargs):
    kwargs['object_hook'] = from_dict
    return json.loads(*args, **kwargs)


def dump(*args, **kwargs):
    kwargs['default'] = to_dict
    return json.dump(*args, **kwargs)


def dumps(*args, **kwargs):
    kwargs['default'] = to_dict
    return json.dumps(*args, **kwargs)



#!/usr/bin/env python

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


with open('jaweson/version.py', 'r') as f:
    exec(f.read())

setup(
    name='jaweson',
    version=__version__,
    description='Powerful serialisation for JSON, without the risk',
    license='BSD',
    author='Adam Griffiths',
    author_email='adam.lw.griffiths@gmail.com',
    url='https://github.com/adamlwgriffiths/jaweson',
    install_requires=[],
    platforms=['any'],
    #test_suite='tests',
    packages=['jaweson'],
    classifiers=[
        'Natural Language :: English',
        'Intended Audience :: Developers',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'License :: OSI Approved :: BSD License',
        'Topic :: Multimedia :: Graphics',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



from __future__ import absolute_import
import inspect
from .serialiser import Serialiser


_types = {}


def register_class(obj):
    global _types
    name = obj.__name__
    if name in _types:
        raise TypeError('A class with the name "{}"" is already defined'.format(name))
    _types[name] = obj


class SerialisableMetaClass(type):
    def __new__(cls, clsname, bases, attrs):
        '''Automatically registers Serialisable subclasses at class definition time.
        '''
        newclass = super(SerialisableMetaClass, cls).__new__(cls, clsname, bases, attrs)
        register_class(newclass)  # here is your register function
        return newclass


class Serialisable(object):
    __metaclass__ = SerialisableMetaClass

    # stores a list of class attributes which should not be serialised
    _blacklist = []

    @classmethod
    def serialisable(cls, key, obj):
        '''Determines what can be serialised and what shouldn't
        '''
        if key.startswith('_'):
            return False
        if key in obj._blacklist:
            return False
        if callable(getattr(obj, key)):
            return False
        # check for properties
        if hasattr(obj.__class__, key):
            if isinstance(getattr(obj.__class__, key), property):
                return False
        return True

    @classmethod
    def to_json(cls, obj):
        '''Serialises the object, by default serialises anything
        that isn't prefixed with _, isn't in the blacklist, and isn't
        callable.
        '''
        return {
            k: getattr(obj, k)
            for k in dir(obj)
            if cls.serialisable(k, obj)
        }
        #raise NotImplementedError('No to_json for {}'.format(cls.__class__.__name__))

    @classmethod
    def from_json(cls, jobj):
        '''Deserialises the object.
        Automatically inspects the object's __init__ function and
        extracts the parameters.
        Can be trivially over-written.
        '''
        try:
            signature = inspect.getargspec(cls.__init__)
            kwargs = {}
            for arg in signature.args:
                if arg in jobj:
                    kwargs[arg] = jobj[arg]

            obj = cls(**kwargs)

            # set any values that aren't passed via the constructor
            # also include the class blacklist in case it has changed
            # we don't want to deserialise newly blacklisted variables
            blacklist = set(['__class__', '__type__'] + cls._blacklist)
            for k in set(jobj.keys()) - set(kwargs.keys()) - blacklist:
                setattr(obj, k, jobj[k])

            # it would be preferable to use this method
            #obj = cls.__new__(cls, **jobj)

            return obj
        except Exception as e:
            raise TypeError('Failed to deserialise {}: {} - args: {}'.format(cls.__name__, str(e), kwargs))


class SerialisableSerialiser(Serialiser):
    python_types = (Serialisable,)
    serialised_types = ('serialisable',)

    def to_json(self, obj):
        if isinstance(obj, Serialisable):
            cls_name = obj.__class__.__name__
            data = obj.to_json(obj)
            data.update({
                '__type__': 'serialisable',
                '__class__': cls_name,
            })
            return data

        return super(SerialisableSerialiser, self).to_json(obj)

    def from_json(self, jobj):
        global _types

        if jobj.get('__type__') == 'serialisable':
            cls_name = jobj['__class__']
            if cls_name not in _types:
                raise NotImplementedError('No type registered for {}'.format(cls_name))

            cls = _types[cls_name]
            if not hasattr(cls, 'from_json'):
                raise NotImplementedError('No from_json classmethod for type {}'.format(cls_name))

            return cls.from_json(jobj)

        return super(SerialisableSerialiser, self).from_json(jobj)
